## ------------------------------------------------------------------------
##
##    $Id: Storage.pm,v 1.16 2003/01/20 16:52:13 joko Exp $
##
##    Copyright (c) 2002  Andreas Motl <andreas.motl@ilo.de>
##
##    See COPYRIGHT section in pod text below for usage and distribution rights.
##
## ------------------------------------------------------------------------
##
##  $Log: Storage.pm,v $
##  Revision 1.16  2003/01/20 16:52:13  joko
##  + now using 'DesignPattern::Object' to create a new 'Data::Storage::Handler::Xyz' on demand - before we did this in a hand-rolled fashion
##
##  Revision 1.15  2003/01/19 03:12:59  joko
##  + modified header
##  - removed pod-documentation - now in 'Storage.pod'
##
##  Revision 1.14  2002/12/19 16:27:59  joko
##  - moved 'sub dropDb' to Data::Storage::Handler::DBI
##
##  Revision 1.13  2002/12/17 21:54:12  joko
##  + feature when using Tangram:
##    + what? each object created should delivered with a globally(!?) unique identifier (GUID) besides the native tangram object id (OID)
##        + patched Tangram::Storage (jonen)
##        + enhanced Data::Storage::Schema::Tangram (joko)
##        + enhanced Data::Storage::Handler::Tangram 'sub getObjectByGuid' (jonen)
##    + how?
##        + each concrete (non-abstract) class gets injected with an additional field/property called 'guid' - this is done (dynamically) on schema level
##        + this property ('guid') gets filled on object creation/insertion from 'sub Tangram::Storage::_insert' using Data::UUID from CPAN
##        + (as for now) this property can get accessed by calling 'getObjectByGuid' on the already known storage-handle used throughout the application
##
##  Revision 1.12  2002/12/12 02:50:15  joko
##  + this now (unfortunately) needs DBI for some helper functions
##  + TODO: these have to be refactored to another scope! (soon!)
##
##  Revision 1.11  2002/12/11 06:53:19  joko
##  + updated pod
##
##  Revision 1.10  2002/12/07 03:37:23  joko
##  + updated pod
##
##  Revision 1.9  2002/12/01 22:15:45  joko
##  - sub createDb: moved to handler
##
##  Revision 1.8  2002/11/29 04:48:23  joko
##  + updated pod
##
##  Revision 1.7  2002/11/17 06:07:18  joko
##  + creating the handler is easier than proposed first - for now :-)
##  + sub testAvailability
##
##  Revision 1.6  2002/11/09 01:04:58  joko
##  + updated pod
##
##  Revision 1.5  2002/10/29 19:24:18  joko
##  - reduced logging
##  + added some pod
##
##  Revision 1.4  2002/10/27 18:35:07  joko
##  + added pod
##
##  Revision 1.3  2002/10/25 11:40:37  joko
##  + enhanced robustness
##  + more logging for debug-levels
##  + sub dropDb
##
##  Revision 1.2  2002/10/17 00:04:29  joko
##  + sub createDb
##  + sub isConnected
##  + bugfixes regarding "deep recursion" stuff
##
##  Revision 1.1  2002/10/10 03:43:12  cvsjoko
##  + new
## ------------------------------------------------------------------------


BEGIN {
  $Data::Storage::VERSION = 0.03;
}

package Data::Storage;

use strict;
use warnings;

use Data::Dumper;
# FIXME: wipe out!
use DBI;

use Data::Storage::Locator;
use DesignPattern::Object;


# TODO: actually implement level (integrate with Log::Dispatch)
my $TRACELEVEL = 0;

# get logger instance
my $logger = Log::Dispatch::Config->instance;

sub new {
  my $invocant = shift;
  my $class = ref($invocant) || $invocant;
  #my @args = normalizeArgs(@_);

  my $arg_locator = shift;
  my $arg_options = shift;

  if (!$arg_locator) {
    $logger->critical( __PACKAGE__ . "->new: No locator passed in!" );
    return;
  }

  #my $self = { STORAGEHANDLE => undef, @_ };
  my $self = { STORAGEHANDLE => undef, locator => $arg_locator, options => $arg_options };
  #$logger->debug( __PACKAGE__ . "[$self->{locator}->{type}]" . "->new(@_)" );
  $logger->debug( __PACKAGE__ . "[$arg_locator->{type}]" . "->new(@_)" );
  return bless $self, $class;
}

sub AUTOLOAD {
  
  # since this is a core function acting as dispatcher to $self->{STORAGEHANDLE},
  # some sophisticated handling and filtering is needed to avoid things like
  #     - Deep recursion on subroutine "Data::Storage::AUTOLOAD"
  #     - Deep recursion on subroutine "Data::Storage::Handler::Abstract::AUTOLOAD"
  #     - Deep recursion on anonymous subroutine at [...]
  # we also might filter log messages caused by logging to itself in "advanced logging of AUTOLOAD calls"
  
  my $self = shift;
  our $AUTOLOAD;
  
  # ->DESTROY would - if not declared - trigger an AUTOLOAD also
  return if $AUTOLOAD =~ m/::DESTROY$/;

  my $method = $AUTOLOAD;
  $method =~ s/^.*:://;

  # advanced logging of AUTOLOAD calls ...
  # ... nice but do it only when TRACING (TODO) is enabled
      my $logstring = "";
      $logstring .= __PACKAGE__ . "[$self->{locator}->{type}]" . "->" . $method;
      #print "count: ", $#_, "\n";
      #$logstring .= Dumper(@_) if ($#_ != -1);
      my $tabcount = int( (80 - length($logstring)) / 10 );
      $logstring .= "\t" x $tabcount . "(AUTOLOAD)";
      # TODO: only ok if logstring doesn't contain 
      #            e.g. "Data::Storage[Tangram]->insert(SystemEvent=HASH(0x5c0034c))          (AUTOLOAD)"
      # but that would be _way_ too specific as long as we don't have an abstract handler for this  ;)
    if ($TRACELEVEL) {
      $logger->debug( $logstring );
      #print join('; ', @_);
    }
    
  # filtering AUTOLOAD calls and first-time-touch of the actual storage impl
  if ($self->_filter_AUTOLOAD($method)) {
    $self->_accessStorage();
    if ($self->{STORAGEHANDLE}) {
      return $self->{STORAGEHANDLE}->$method(@_);
    } else {
      $logger->critical( __PACKAGE__ . "->AUTOLOAD: ERROR: " . $logstring );
      return;
    }
  }
  
}

sub _filter_AUTOLOAD {
  my $self = shift;
  my $method = shift;
  if ($self->{options}->{protected}) {
    if ($method eq 'disconnect') {
      return;
    }
  }
  return 1;
}


sub normalizeArgs {
  my %args = @_;
  if (!$args{dsn} && $args{meta}{dsn}) {
    $args{dsn} = $args{meta}{dsn};
  }
  my @result = %args;
  return @result;
}

sub _accessStorage {
  my $self = shift;
  # TODO: to some tracelevel!
  if ($TRACELEVEL) {
    $logger->debug( __PACKAGE__ .  "[$self->{locator}->{type}]" . "->_accessStorage()" );
  }
  if (!$self->{STORAGEHANDLE}) {
    return $self->_createStorageHandle();
  }
}

sub _createStorageHandle1 {
  my $self = shift;
  my $type = $self->{locator}->{type};
  $logger->debug( __PACKAGE__ .  "[$type]" . "->_createStorageHandle()" );

  my $pkg = "Data::Storage::Handler::" . $type . "";
  
  # try to load perl module at runtime
  my $evalstr = "use $pkg;";
  eval($evalstr);
  if ($@) {
    $logger->error( __PACKAGE__ .  "[$type]" . "->_createStorageHandle(): $@" );
    return;
  }
  
  # build up some additional arguments to pass on
  #my @args = %{$self->{locator}};
  my @args = ();

  # - create new storage handle object
  # - propagate arguments to handler
  # - pass locator by reference to be able to store status- or meta-information in it
  $self->{STORAGEHANDLE} = $pkg->new( locator => $self->{locator}, @args );

}

sub _createStorageHandle {
  my $self = shift;
  my $type = $self->{locator}->{type};
  $logger->debug( __PACKAGE__ .  "[$type]" . "->_createStorageHandle()" );

#print Dumper($self);
#exit;

  my $pkg = "Data::Storage::Handler::" . $type . "";
  
  # try to load perl module at runtime
=pod
  my $evalstr = "use $pkg;";
  eval($evalstr);
  if ($@) {
    $logger->error( __PACKAGE__ .  "[$type]" . "->_createStorageHandle(): $@" );
    return;
  }
  
  # build up some additional arguments to pass on
  #my @args = %{$self->{locator}};
  my @args = ();

  # - create new storage handle object
  # - propagate arguments to handler
  # - pass locator by reference to be able to store status- or meta-information in it
  $self->{STORAGEHANDLE} = $pkg->new( locator => $self->{locator}, @args );
=cut

  $self->{STORAGEHANDLE} = DesignPattern::Object->fromPackage( $pkg, locator => $self->{locator} );
  return 1;

}

sub addLogDispatchHandler {

      my $self = shift;
      my $name = shift;
      my $package = shift;
      my $logger1 = shift;
      my $objectCreator = shift;
      
      #$logger->add( Log::Dispatch::Tangram->new( name => $name,
      $logger->add( $package->new( name => $name,
                                            #min_level => 'debug',
                                            min_level => 'info',
                                            storage => $self,
                                            objectCreator => $objectCreator,
                                            fields => {
                                              message => 'usermsg',
                                              timestamp => 'stamp',
                                              level => 'level',
                                              name => 'code',
                                            },
                                            filter_patterns => [ '->insert\(SystemEvent=' ],
                                            #filter_patterns => [ 'SystemEvent' ],
                                            
                                            #format  => '[%d] [%p] %m%n',
                                             ) );

}

sub removeLogDispatchHandler {
  my $self = shift;
  my $name = shift;
  #my $logger = shift;
  $logger->remove($name);
}

=pod
sub getDbName {
  my $self = shift;
  my $dsn = $self->{locator}->{dbi}->{dsn};
  $dsn =~ m/database=(.+?);/;
  my $database_name = $1;
  return $database_name;
}

sub testAvailability {
  my $self = shift;
  my $status = $self->testDsn();
  $self->{locator}->{status}->{available} = $status;
  return $status;
}

sub isConnected {
  my $self = shift;
  # TODO: REVIEW!
  return 1 if $self->{STORAGEHANDLE};
}

sub testDsn {
  my $self = shift;
  my $dsn = $self->{locator}->{dbi}->{dsn};
  my $result;
  if ( my $dbh = DBI->connect($dsn, '', '', {
                                                      PrintError => 0,
                                                      } ) ) {
    
    # TODO: REVIEW
    $dbh->disconnect();
    
    return 1;
  } else {
    $logger->warning( __PACKAGE__ .  "[$self->{locator}->{type}]" . "->testDsn(): " . "DBI-error: " . $DBI::errstr );
  }
}
=cut

1;
__END__
